######################################


#1D- double adiabatic MHD functions


######################################
import numpy as np
from scipy.special import erfc
import CGLMHD
def Generate_Grid_1D_uniform(nx,NO):
    #nx - number of cell corners in the x direction
    xmin=-1.
    xmax=1.
    NO2=NO/2
    dx = (xmax-xmin)/nx
    x = np.arange((xmin-NO2*dx),(xmax+(NO2+1)*dx),dx)
    return (x)
def getConservedVariables(rho,vx,vy,vz,p,gamma):
    """
    Function computes the conserved form of the plasma parameters 
    Requries:
        rho,vx,vy,vz,p - plasma variables
        gamma - ratio of specific heats
    Returns:
        rho,rhovx,rhovy,rhovz,eng - plasma variables in conserved form
    """
    rho = rho
    rhovx = rho*vx
    rhovy = rho*vy
    rhovz = rho*vz
    eng= 0.5*rho*(vx**2+vy**2+vz**2)+p/(gamma-1)
    
    return (rho,rhovx,rhovy,rhovz,eng)

def getAdiabaticinvariants (rho,pper,ppar,bx,by,bz):
    '''
    The first adiabatic invariant s1 is magnetic moment, mu=pper/b
    The second adiabatic invariant s2 is ppar*b**2/rho**2
    '''
    bm = np.sqrt(bx**2+by**2+bz**2)
    s1 = pper/bm
    s2 = ppar*bm**2/rho**2
    return (s1,s2)


def getDT(rho,vx,vy,vz,bx,by,bz,p,gamma,dx,dy,dz,CFL):
    """
    Function computes the timestep based upon computed wave speeds and the 
    CFL condition 
    Requries:
        rho,vx,vy,vz,bx,by,bz,p - plasma variables and magnetic field vars
        gamma - ratio of specific heats
        dx,dy,dz - size of edges
        CFL - CFL condition parameter
    Returns:
        dt - smallest allowed timestep
    """               

    
    Vfluid = np.sqrt(vx**2+vy**2+vz**2)
    Btotal = np.sqrt(bx**2+by**2+bz**2)
    Valfvn = Btotal/np.sqrt(rho)
    Vsound = np.sqrt(gamma*p/rho)
    
    VCFL = Vfluid+np.sqrt(Valfvn**2+Vsound**2)
    dtCFL = CFL /(VCFL/dx+VCFL/dy+VCFL/dz)
    dt = dtCFL.min()
    
    return (dt)

"""
Compute return sign of number with zero replaced as 1
"""
def sign1(x):
    """
    Emulates fortran sign function by returning 1 for zero locations 
    Requries:
        x - array to check for signs
    Returns:
        a - array same size x with sign values
    """  

    
    a = np.sign(x)
    a[a == 0] = 1
    
    return a
    
def WENO5(hm2,hm1,h,hp1,hp2):
    """
    Function compoutes the left and right PDM limited states 
    Requries:
        hm2 - ?
        hm1 - ?
        h - ?
        hp1 - ?
        hp2 - ?
    Returns:
        data - WENO recostruction
    """  



    ## WENO-5 reconstruction
    h0 = 1./3.*hm2-7/6*hm1+11./6.*h
    h1 = -1./6.*hm1+5/6*h+1./3.*hp1
    h2 = 1./3.*h+5/6*hp1-1./6.*hp2
    
    IS0 = 13./12.*(hm2 - 2*hm1 + h)**2 + 1./4.*(hm2 - 4*hm1 + 3*h)**2
    IS1 = 13./12.*(hm1 - 2*h + hp1)**2 + 1./4.*(hm1 - hp1)**2
    IS2 = 13./12.*(h - 2*hp1 + hp2)**2 + 1./4.*(3*h - 4*hp1 + hp2)**2
    epsilon = 1.0e-8
    
    a0 = 1./10./(epsilon+IS0)**2
    a1 = 6./10./(epsilon+IS1)**2
    a2 = 3./10./(epsilon+IS2)**2
    
    
    w0 = a0/(a0+a1+a2)
    w1 = a1/(a0+a1+a2)
    w2 = a2/(a0+a1+a2)
    
    data = w0*h0 + w1*h1 + w2*h2
    
    return data

"""
Compute left and right PDM states
"""
def PDM2(f0,f1,f2,f3,f,PDMB):
    """
    Function compoutes the left and right PDM limited states 
    Requries:
        f0 - ?
        f1 - ?
        f2 - ?
        f3 - ?
        f - ?
        PDMB - value of the B parameter in the limiter 
    Returns:
        (f_left,f_right) - Limited left/right state values
    """  

    ##
    # f0=1
    # f1=1
    # f2=1
    # f3=0
    
    # second order interp
    # f=0.5*(f1+f2) 
    
    # fourth order interp
    # f = (-f0+9*f1+9*f2-f3)/16
    # f = (-f0+7*f1+7*f2-f3)/12
    
    #NB - the maximum and minimum functions do an element by element comparison
    #     of the maxtrix elements and return the corresponding max/min
    maxf = np.maximum(f1,f2)
    minf = np.minimum(f1,f2)
    
    f = np.maximum(minf,np.minimum(f,maxf))
    
    df0 = PDMB*(f1-f0)
    df1 = PDMB*(f2-f1)
    df2 = PDMB*(f3-f2)
    
    s0 = CGLMHD.sign1(df0)
    s1 = CGLMHD.sign1(df1)
    s2 = CGLMHD.sign1(df2)
    
    df0 = np.abs(df0)
    df1 = np.abs(df1)
    df2 = np.abs(df2)
    
    q0 = np.abs(s0+s1)
    q1 = np.abs(s1+s2)
    
    df_left = f - f1
    df_righ = f2 - f
    
    f_left = f - s1*np.maximum(0,np.abs(df_left) - q0*df0)
    f_right= f + s1*np.maximum(0,np.abs(df_righ) - q1*df2) 
    
    return (f_left,f_right)
"""
Compute the left and right states
"""

"""
Compute left and right TVD states
"""
def TVD(fm1,f0,f1,f2):
    """
    Function compoutes the left and right PDM limited states 
    Requries:
        fm1 - ?
        f1 - ?
        f2 - ?
 
    Returns:
        (f_left,f_right) - Limited left/right state values
    """     
    # left value
    eps=1.0e-10
    r_i = (f0-fm1)/(f1-f0+eps)
    phi_i = (r_i + abs(r_i))/(1+abs(r_i)) # Van Leer slope limiter
    # phi_i = max(0,max(min(2.*r_i,1),min(r_i,2))) # Superbee slope limiter
    
    f = 0.5*(f0+f1)
    f_left = f0 - phi_i*(f0-f)
    
    # right value
    r_i = (f1-f2)/(f0-f1+eps)
    phi_i = (r_i + abs(r_i))/(1+abs(r_i))
    # phi_i = max(0,max(min(2.*r_i,1),min(r_i,2)))
    
    f = 0.5*(f1+f0)
    f_right = f1 - phi_i*(f1-f)
    
    return (f_left,f_right)

def reconstruct_1D_x(rho_h,NO2,PDMB,limiter_type='PDM'):
    """
    Function compoutes the left and right states at the 
    Requries:
        rho_h - parameter to reconstruct
        NO2 - Half numerical order
        PDMB - B parameter for PDM method
        direction - 1 for x, 2 for y, and 3 for z
        limiter_type - limiter_type of limiter to use in 
    Returns:
        rho_left,rho_right - left and right states
    """               

    
    #8-th order reconstruction
    
    if(limiter_type == 'PDM'):
        rho_interp = (
                    -3*rho_h[NO2-4:-NO2-4+1]+
                    29*rho_h[NO2-3:-NO2-3+1]-
                    139*rho_h[NO2-2:-NO2-2+1]+
                    533*rho_h[NO2-1:-NO2-1+1]+
                    533*rho_h[NO2:-NO2+1]-
                    139*rho_h[NO2+1:-NO2+1+1]+
                    29*rho_h[NO2+2:-NO2+2+1]-
                    3*rho_h[NO2+3:])/840
        
        (rho_left,rho_right)= CGLMHD.PDM2(
                                rho_h[NO2-2:-NO2-2+1],
                                rho_h[NO2-1:-NO2-1+1],
                                rho_h[NO2:-NO2+1    ],
                                rho_h[NO2+1:-NO2+1+1],
                                rho_interp,PDMB)
    elif(limiter_type == 'TVD'):
        (rho_left,rho_right)= CGLMHD.TVD(
                                rho_h[NO2-2:-NO2-2+1],
                                rho_h[NO2-1:-NO2-1+1],
                                rho_h[NO2:-NO2+1],
                                rho_h[NO2+1:-NO2+1+1])
    elif(limiter_type == 'WENO'):
        rho_left = CGLMHD.WENO5(
                                rho_h[NO2-3:-NO2-3+1],
                                rho_h[NO2-2:-NO2-2+1],
                                rho_h[NO2-1:-NO2-1+1],
                                rho_h[NO2:-NO2+1],
                                rho_h[NO2+1:-NO2+1+1])
        rho_right = CGLMHD.WENO5(
                                rho_h[NO2+2:-NO2+2+1],
                                rho_h[NO2+1:-NO2+1+1],
                                rho_h[NO2:-NO2+1],
                                rho_h[NO2-1:-NO2-1+1],
                                rho_h[NO2-2:-NO2-2+1])

    
    return (rho_left,rho_right)

"""
Compute the conserved variables and return them in a tuple
"""
def getHydroFlux(rho,Vx,Vy,Vz,p,pper,bx,by,bz):
    """
    Function computes the conserved form of the plasma parameters 
    Requries:
        rho,vx,vy,vz,p - plasma variables

    Returns:
        (Frho_p,FrhoVx_p,FrhoVy_p,FrhoVz_p,Feng_p,Fs1_p,Fs2_p
         Frho_n,FrhoVx_n,FrhoVy_n,FrhoVz_n,Feng_n,Fs1_n,Fs2_n) - Tuple of fluxes
    """


    # This function calculates the positive/negative moments of a Bi-maxwellian 
    # distribution for the face flux calculation.
    
    pi = np.pi
    a = np.sqrt(pper/rho)
    ppar = 3.*p-2.*pper

    b = np.sqrt(ppar/rho) 
    theta = np.arctan(by/bx)
    eta = np.sqrt(a**2.*np.sin(theta)**2+b**2.*np.cos(theta)**2)
    
    qerfc = erfc(-Vx/(2**0.5*eta))
    #quick erfc variable for later use
    qexp = np.exp(-Vx**2./(2.*eta**2))
    #quick exp variable for later use
    bm = np.sqrt(bx**2+by**2+bz**2)
    #magnetic field magnitude
    
    
    #integrals needed for flux calculations
    I00_p = qerfc/2
    I10_p = eta*qexp/np.sqrt(2* pi)
    I20_p = (eta**2 *qerfc-np.sqrt(2./pi)*eta*Vx*qexp)/2
    I30_p = eta*(2*eta**2+Vx**2)*qexp/np.sqrt(2.*pi)
    I01_p = (b**2-a**2)*np.sin(theta)*np.cos(theta)*qexp/(np.sqrt(2.*pi)*eta)
    I11_p = 0.5*(b**2-a**2)*np.sin(theta)*np.cos(theta)*(qerfc-np.sqrt(2./pi)*Vx*qexp/eta)
    I21_p = np.sin(theta)*np.cos(theta)*(b**2-a**2)*(2.*eta**2+Vx**2)*qexp/(np.sqrt(2.*pi)*eta)
    I02_p = (-1.*(b**2-a**2)**2*np.cos(theta)**2.*np.sin(theta)**2.*Vx*qexp/np.sqrt(2.*pi) \
             +eta/2.*qerfc*(np.sin(theta)**2.*np.cos(theta)**2.*(b**2-a**2)**2+a**2.*b**2))/eta**3           
    I12_p = qexp*(a**2.*b**2.*eta**2\
                  +np.sin(theta)**2.*np.cos(theta)**2.*(2.*eta**2+Vx**2)*(a**2-b**2)**2)/(np.sqrt(2.*pi)*eta**3)                 
    
    I00 = 1 
    I10 = 0 
    I20 = eta**2 
    I30 = 0 
    I01 = 0 
    I11 = (b**2-a**2)*np.cos(theta)*np.sin(theta) 
    I21 = 0 
    I02 = a**2+(b**2-a**2)*np.sin(theta)**2
    I12 = 0 
    
    I00_n = I00-I00_p
    I10_n = I10-I10_p
    I20_n = I20-I20_p
    I30_n = I30-I30_p
    I01_n = I01-I01_p
    I11_n = I11-I11_p
    I21_n = I21-I21_p
    I02_n = I02-I02_p
    I12_n = I12-I12_p

        
    # Fluxes in the positive direction
    Frho_p   = rho  * (Vx*I00_p+I10_p)
    FrhoVx_p = rho*(Vx**2.*I00_p+2.*Vx*I10_p+I20_p)
    FrhoVy_p = rho*(Vx*Vy*I00_p+Vx*I01_p+Vy*I10_p+I11_p)
    FrhoVz_p = rho*(Vx*Vz*I00_p+Vz*I10_p)

    Feng_p   = 0.5*rho*(Vx*(a**2+Vx**2+Vy**2+Vz**2)*I00_p+(a**2+3.*Vx**2+Vy**2+Vz**2)*I10_p\
                         +3.*Vx*I20_p+I30_p+2.*Vx*Vy*I01_p+Vx*I02_p+2*Vy*I11_p+I12_p)
        
    Fs1_p = 0.5*rho/bm*(Vx*(I00_p*a**2+I02_p*np.cos(theta)**2-2.*I11_p*np.cos(theta)*np.sin(theta)+I20_p*np.sin(theta)**2)\
                    +I10_p*a**2+I12_p*np.cos(theta)**2-2.*I21_p*np.cos(theta)*np.sin(theta)+I30_p*np.sin(theta)**2)
                           
    #s1 stands for first adiabatic invariant mu(magnetic moment)
    Fs2_p = bm**2./rho*(Vx*(I02_p*np.sin(theta)**2+2.*I11_p*np.sin(theta)*np.cos(theta)+I20_p*np.cos(theta)**2)\
                         +I12_p*np.sin(theta)**2+2.*I21_p*np.cos(theta)*np.sin(theta)+I30_p*np.cos(theta)**2)
    #s2 stands for the second adiabatic invariant
    
    
    # Fluxes in the negative direction
    Frho_n   = rho  * (Vx*I00_n+I10_n)
    FrhoVx_n = rho*(Vx**2.*I00_n+2.*Vx*I10_n+I20_n)
    FrhoVy_n = rho*(Vx*Vy*I00_n+Vx*I01_n+Vy*I10_n+I11_n)
    FrhoVz_n = rho*(Vx*Vz*I00_n+Vz*I10_n)

    Feng_n   = 0.5*rho*(Vx*(a**2+Vx**2+Vy**2+Vz**2)*I00_n+(a**2+3.*Vx**2+Vy**2+Vz**2)*I10_n\
                         +3.*Vx*I20_n+I30_n+2.*Vx*Vy*I01_n+Vx*I02_n+2*Vy*I11_n+I12_n)
        
    Fs1_n = 0.5*rho/bm*(Vx*(I00_n*a**2+I02_n*np.cos(theta)**2-2.*I11_n*np.cos(theta)*np.sin(theta)+I20_n*np.sin(theta)**2)\
                    +I10_n*a**2+I12_n*np.cos(theta)**2-2.*I21_n*np.cos(theta)*np.sin(theta)+I30_n*np.sin(theta)**2)
                           
    #s1 stands for first adiabatic invariant mu(magnetic moment)
    Fs2_n = bm**2./rho*(Vx*(I02_n*np.sin(theta)**2+2.*I11_n*np.sin(theta)*np.cos(theta)+I20_n*np.cos(theta)**2)\
                         +I12_n*np.sin(theta)**2+2.*I21_n*np.cos(theta)*np.sin(theta)+I30_n*np.cos(theta)**2)
        
            
    return (Frho_p,FrhoVx_p,FrhoVy_p,FrhoVz_p,Feng_p,Fs1_p,Fs2_p,
            Frho_n,FrhoVx_n,FrhoVy_n,FrhoVz_n,Feng_n,Fs1_n,Fs2_n) 
 

def getMagneticStress(rho,Vx,p,pper,bx,by,bz):
    
    #This function calculates the positive/negative moments of a Bi-maxwellian
    # distribution for the magnetic stress calculation.
    #B field is included in the total P to have the mean speed within the
    # distribution equal to the fast mode speed.
    
    pi = np.pi
    bm = np.sqrt(bx**2+by**2+bz**2)
    #magnetic field magnitude
    
    pper_tot = pper+ 0.5*bm**2
    ppar = 3.*p-2.*pper
    ppar_tot = ppar+ 0.5*bm**2
    a = np.sqrt(pper_tot/rho)
    b = np.sqrt(ppar_tot/rho)
    
    
    theta = np.arctan(by/bx)
    eta = np.sqrt(a**2.*np.sin(theta)**2+b**2.*np.cos(theta)**2)
    
    qerfc = erfc(-Vx/(2**0.5*eta))
    #quick erfc variable for later use
    qexp = np.exp(-Vx**2./(2.*eta**2))
    #quick exp variable for later use
    
    
    I00_p = qerfc/2
    I00 = 1
    I00_n = I00-I00_p
    
#   Flux in the positive x direction

    Bstress_x_p = (0.5*bm**2 - bx*bx) * I00_p
    Bstress_y_p = (0- by*bx) * I00_p
    Bstress_z_p = (0 - bz*bx) * I00_p
    
#  Flux in the negative x direction

    
    Bstress_x_n = (0.5*bm**2 - bx*bx) * I00_n
    Bstress_y_n = (0- by*bx) * I00_n
    Bstress_z_n = (0 - bz*bx) * I00_n
    
    return (Bstress_x_p,Bstress_y_p,Bstress_z_p,
            Bstress_x_n,Bstress_y_n,Bstress_z_n)

def getMagneticFlux(rho,Vx,Vy,p,pper,bx,by,bz):
    
    
    
    pi = np.pi
    bm = np.sqrt(bx**2+by**2+bz**2)
    #magnetic field magnitude
    
    pper_tot = pper+ 0.5*bm**2
    ppar = 3.*p-2.*pper
    ppar_tot = ppar+ 0.5*bm**2
    a = np.sqrt(pper_tot/rho)
    b = np.sqrt(ppar_tot/rho)
    
    
    theta = np.arctan(by/bx)
    eta = np.sqrt(a**2.*np.sin(theta)**2+b**2.*np.cos(theta)**2)
    
    qerfc = erfc(-Vx/(2**0.5*eta))
    #quick erfc variable for later use
    qexp = np.exp(-Vx**2./(2.*eta**2))
    #quick exp variable for later use
    
    
    I00_p = qerfc/2
    I00 = 1
    I00_n = I00-I00_p
    
    I10_p = eta*qexp/np.sqrt(2* pi)
    I10 = 0
    I10_n = I10-I10_p
    
#   Flux in the positive x direction

    Fbx_p    = 0*bx
    Fby_p    = by* (I10_p+Vx*I00_p) - Vy* bx* I00_p
    Fbz_p    = 0*bz
    
#  Flux in the negative x direction

    
    Fbx_n    = 0*bx
    Fby_n    = by* (I10_n+Vx*I00_n) - Vy* bx * I00_n
    Fbz_n    = 0*bz
    
    return (Fbx_p,Fby_p,Fbz_p, 
           Fbx_n,Fby_n,Fbz_n)