#################################################################
#        1-D Double Adiabatic MHD solver on Cartesian Grid 
#
#   
#    FEATURES:
#    1. Finite Volume(in 1-D, it's equivalant to finite difference)
#    2. TVD/TVB limiter - PDM/WENO
#    2. Semi-conservative - only track plasma energy
#    3. high order reconstruction in space
#    4. 2nd order Adam-Bashforth scheme in time
#    5. Operator-splitting for the Lorentz force
#    6. Beam scheme for plasma and magnetic field
#
#
##################################################################

import numpy as np
import pylab as pl
import CGLMHD
import time,os
import matplotlib.pyplot as plt
import copy
import matplotlib.colors as mcol
import matplotlib.cm as cm

  
    
#  Model Parameters
NO = int(8) #order of reconstruction
NO2 = int(NO/2) # number of ghost cells on each side
gamma=5/3
CFL = 0.3
PDMB = 0.25

limiter_type = 'PDM'  # 'PDM' - 8th order with PDM limiter
                     # 'TVD' - 2nd order with Van-Leer limiter
                      # 'WENO'- 5th order with WENO reconstruction 



# Grid information- nx is # of ACTIVE cells (no ghost cell included)
# The generated grid are cell corners, the metric function will calculate
# cell centers, faces and other grid information
nx = 512

x=CGLMHD.Generate_Grid_1D_uniform(nx,NO) # This function generate a 
                                               # uniformly distributed active 
                                               # grid between -1 and 1 with 
                                               # nx cells in each 
                                               # direction

xc = 0.5*(x[:-1]+x[1:]) #cell center grid
dx = x[1:] - x[:-1]    #dx of each cell

# Define premitive Hydrodynamic variables at cell center
rho = np.zeros(xc.shape)
vx = np.zeros(xc.shape)
vy = np.zeros(xc.shape)
vz = np.zeros(xc.shape)
p = np.zeros(xc.shape)
#p perpendicular and p parallel 
pper = np.zeros(xc.shape)
ppar = np.zeros(xc.shape)

# Define magnetic fields at cell center - for 1-D MHD, the cell centered
# bx, by, bz are the evolved variables as long as bx=const. (which satisfies div B = 0)
bx = np.zeros(xc.shape)
by = np.zeros(xc.shape)
bz = np.zeros(xc.shape)


# 1-D Brio-Wu shock tube problem
rho[xc<0.0] = 1.0
rho[xc>0.0] = 0.125
p[xc<0.0]   = 1.0
p[xc>0.0]   = 0.1
bx[:]  =  0.75
by[xc<0.0]  = 1.0
by[xc>0.0]  = -1.0
pper = 1.*p
ppar = 3.*p-2.*pper


#Outflow boundary
#:NO2 gives the left boundary cells
# NO-1:NO2-1:-1 gives the first NO2 active cells in reverse order      
rho[:NO2] =rho[NO-1:NO2-1:-1]
vx[:NO2]  = vx[NO-1:NO2-1:-1]
vy[:NO2]  = vy[NO-1:NO2-1:-1]
vz[:NO2]  = vz[NO-1:NO2-1:-1]
p[:NO2]   =  p[NO-1:NO2-1:-1]

by[:NO2]  = by[NO-1:NO2-1:-1]
bz[:NO2]  = bz[NO-1:NO2-1:-1]
pper[:NO2]   =  pper[NO-1:NO2-1:-1]
# ppar[:NO2]   =  ppar[NO-1:NO2-1:-1]


#-NO2: gives the right boundary cells
#-NO2-1:-NO-1:-1 gives last NO2 active cells in reverse order
rho[-NO2:] = rho[-NO2-1:-NO-1:-1]
p[-NO2:] = p[-NO2-1:-NO-1:-1]
vx[-NO2:] = vx[-NO2-1:-NO-1:-1]
vy[-NO2:] = vy[-NO2-1:-NO-1:-1]
vz[-NO2:] = vz[-NO2-1:-NO-1:-1]

by[-NO2:] = by[-NO2-1:-NO-1:-1] 
bz[-NO2:] = bz[-NO2-1:-NO-1:-1]
pper[-NO2:] = pper[-NO2-1:-NO-1:-1]
# ppar[-NO2:] = ppar[-NO2-1:-NO-1:-1]


(rho,rhovx,rhovy,rhovz,eng) = CGLMHD.getConservedVariables(rho,vx,vy,vz,p,gamma)
(s1,s2) = CGLMHD.getAdiabaticinvariants(rho,pper,ppar,bx,by,bz)

dy=1e20
dz=1e20
dt0 = CGLMHD.getDT(rho,vx,vy,vz,bx,by,bz,p,gamma,dx,dy,dz,CFL)


# Save the initial states for the first Adam-Bashforth time stepping
rho_p = rho
vx_p = vx
vy_p = vy
vz_p = vz
p_p= p
bx_p = bx
by_p = by
bz_p = bz
pper_p= pper
ppar_p= ppar

# MAIN LOOP
# simulation Time information
count = 0
Nstep=1e6
Time = 0
RealT=0
step=0


for step in np.arange(Nstep):
    
    dt = CGLMHD.getDT(rho,vx,vy,vz,bx,by,bz,p,gamma,dx,dy,dz,CFL)
    Time = Time + dt
    [rho,rhovx,rhovy,rhovz,eng] = CGLMHD.getConservedVariables(rho,vx,vy,vz,p,gamma)
    (s1,s2) = CGLMHD.getAdiabaticinvariants(rho,pper,ppar,bx,by,bz)
     
    
    # Step 1: get the half time step values
    rho_h = rho + dt/dt0/2*(rho-rho_p)
    vx_h = vx + dt/dt0/2*(vx-vx_p)
    vy_h = vy + dt/dt0/2*(vy-vy_p)
    vz_h = vz + dt/dt0/2*(vz-vz_p)
    p_h = p + dt/dt0/2*(p-p_p)
    bx_h = bx + dt/dt0/2*(bx-bx_p)
    by_h = by + dt/dt0/2*(by-by_p)
    bz_h = bz + dt/dt0/2*(bz-bz_p)
    p_h = p + dt/dt0/2*(p-p_p)
    pper_h = pper + dt/dt0/2*(pper-pper_p)
    ppar_h = ppar + dt/dt0/2*(ppar-ppar_p)
    
    rho0 = rho
    rhovx0 = rhovx
    rhovy0 = rhovy
    rhovz0 = rhovz
    by0 = by
    bz0 = bz
    eng0 = eng
    s10 = s1
    
    # save the current state vector for next AB time stepping
    rho_p = rho
    vx_p = vx
    vy_p = vy
    vz_p = vz
    bx_p = bx
    by_p = by
    bz_p = bz
    p_p= p
    pper_p= pper
    ppar_p= ppar
    dt0 = dt
    
    # step 2: reconstruct the primitive variables to cell faces
    (rho_left, rho_right) = CGLMHD.reconstruct_1D_x(rho_h,NO2,PDMB,limiter_type)    
    (vx_left, vx_right) = CGLMHD.reconstruct_1D_x(vx_h,NO2,PDMB,limiter_type)   
    (vy_left, vy_right) = CGLMHD.reconstruct_1D_x(vy_h,NO2,PDMB,limiter_type)       
    (vz_left, vz_right) = CGLMHD.reconstruct_1D_x(vz_h,NO2,PDMB,limiter_type)
    (bx_left, bx_right) = CGLMHD.reconstruct_1D_x(bx_h,NO2,PDMB,limiter_type)
    (by_left, by_right) = CGLMHD.reconstruct_1D_x(by_h,NO2,PDMB,limiter_type)
    (bz_left, bz_right) = CGLMHD.reconstruct_1D_x(bz_h,NO2,PDMB,limiter_type)
    (p_left, p_right) = CGLMHD.reconstruct_1D_x(p_h,NO2,PDMB,limiter_type)

    (pper_left, pper_right) = CGLMHD.reconstruct_1D_x(pper_h,NO2,PDMB,limiter_type)
    (ppar_left, ppar_right) = CGLMHD.reconstruct_1D_x(ppar_h,NO2,PDMB,limiter_type)
    
    b_left = np.sqrt(bx_left**2+by_left**2+bz_left**2)
    b_right = np.sqrt(bx_right**2+by_right**2+bz_right**2)
    
#     step 3: calculate the hydro/magnetic flux/stress through cell faces
    (Frho_p,FrhoVx_p,FrhoVy_p,FrhoVz_p,Feng_p,Fs1_p,Fs2_p,_,_,_,_,_,_,_) = CGLMHD.getHydroFlux(
                    rho_left,vx_left,vy_left,vz_left,p_left,pper_left,bx_left,by_left,bz_left)
    (_,_,_,_,_,_,_,Frho_n,FrhoVx_n,FrhoVy_n,FrhoVz_n,Feng_n,Fs1_n,Fs2_n) = CGLMHD.getHydroFlux(
                    rho_right,vx_right,vy_right,vz_right,p_right,pper_right,bx_right,by_right,bz_right)
    
    rho_flux_x = Frho_p + Frho_n
    vx_flux_x = FrhoVx_p + FrhoVx_n
    vy_flux_x = FrhoVy_p + FrhoVy_n
    vz_flux_x = FrhoVz_p + FrhoVz_n
    eng_flux_x = Feng_p + Feng_n
    s1_flux_x = Fs1_p + Fs1_n
    

    
#     Add hogs diffusion
#     compute alfven velocity
    Va_eff = np.sqrt((bx**2+by**2+bz**2)/(rho))
    dif_factor = 0.5
    diff_va = 0*Va_eff

    
    diff_va[NO2:-NO2+1] = dif_factor*(Va_eff[NO2-1:-NO2+1-1] + Va_eff[NO2:-NO2+1])/2
    rho_flux_x[:] = rho_flux_x[:] - diff_va[NO2:-NO2+1]*(rho_right[:] - rho_left[:])
    vx_flux_x[:]  = vx_flux_x[:]  - diff_va[NO2:-NO2+1]*(rho_right[:]*vx_right[:] - rho_left[:]*vx_left[:])
    vy_flux_x[:]  = vy_flux_x[:]  - diff_va[NO2:-NO2+1]*(rho_right[:]*vy_right[:] - rho_left[:]*vy_left[:])
    vz_flux_x[:]  = vz_flux_x[:]  - diff_va[NO2:-NO2+1]*(rho_right[:]*vz_right[:] - rho_left[:]*vz_left[:])
    eng_flux_x[:]= eng_flux_x[:] - diff_va[NO2:-NO2+1]*(0.5*rho_right[:]*(vx_right[:]**2+vy_right[:]**2+vz_right[:]**2) + p_right[:]/(gamma-1)\
                                                        - 0.5*rho_left[:]*(vx_left[:]**2+vy_left[:]**2+vz_left[:]**2) - p_left[:]/(gamma-1))
                                                                   
    s1_flux_x[:]= s1_flux_x[:]  - diff_va[NO2:-NO2+1]*(pper_right[:]/b_right[:]-pper_left[:]/b_left[:])

                                                  
    
    # Magnetic distribution function is also Bi-maxwellian, the "temperature" is
    # fluid+magnetic
    (Bstress_x_p, Bstress_y_p, Bstress_z_p,_, _,_) = CGLMHD.getMagneticStress(
        rho_left,vx_left,p_left,pper_left,bx_left,by_left,bz_left)
    (_,_,_,Bstress_x_n, Bstress_y_n, Bstress_z_n) = CGLMHD.getMagneticStress(
        rho_right,vx_right,p_right,pper_right,bx_right,by_right,bz_right)
    
    BstressX_x = Bstress_x_p + Bstress_x_n
    BstressY_x = Bstress_y_p + Bstress_y_n
    BstressZ_x = Bstress_z_p + Bstress_z_n

    
#   maxwell's equation
    (Fbx_p,Fby_p,Fbz_p, _,_,_) = CGLMHD.getMagneticFlux(rho_left,vx_left,vy_left,p_left,pper_left,bx_left,by_left,bz_left)
    
    (_,_,_,Fbx_n,Fby_n,Fbz_n) = CGLMHD.getMagneticFlux(rho_right,vx_right,vy_right,p_right,pper_right,bx_right,by_right,bz_right)
    
    by_flux_x = Fby_p + Fby_n
    bz_flux_x = Fbz_p + Fbz_n
    
#      step 4, update conserved fluid variables using finite-volume method, to get the correct rho and p, note
#      that in 1-D, it's equivalant to finite difference
    rho[NO2:-NO2] = rho0[NO2:-NO2] - dt*(rho_flux_x[1:]-rho_flux_x[:-1])/dx[NO2:-NO2]
    rhovx[NO2:-NO2] = rhovx0[NO2:-NO2] - dt*(vx_flux_x[1:]-vx_flux_x[:-1])/dx[NO2:-NO2]
    rhovy[NO2:-NO2]= rhovy0[NO2:-NO2] - dt*(vy_flux_x[1:]-vy_flux_x[:-1])/dx[NO2:-NO2]
    rhovz[NO2:-NO2] = rhovz0[NO2:-NO2] - dt*(vz_flux_x[1:]-vz_flux_x[:-1])/dx[NO2:-NO2]
    eng[NO2:-NO2] = eng0[NO2:-NO2] - dt*(eng_flux_x[1:]-eng_flux_x[:-1])/dx[NO2:-NO2]
    s1[NO2:-NO2] = s10[NO2:-NO2]  - dt*(s1_flux_x[1:]-s1_flux_x[:-1])/dx[NO2:-NO2]
    vx = rhovx/rho
    vy = rhovy/rho
    vz = rhovz/rho
    p = (eng - 0.5*rho*(vx**2+vy**2+vz**2))*(gamma-1)
    


#     then apply the magnetic stress to get the correct velocity
    rhovx[NO2:-NO2] = rhovx[NO2:-NO2]  - dt/dx[NO2:-NO2]*(BstressX_x[1:]-BstressX_x[:-1]) 
    rhovy[NO2:-NO2]= rhovy[NO2:-NO2] - dt/dx[NO2:-NO2]*(BstressY_x[1:]-BstressY_x[:-1])
    rhovz[NO2:-NO2] = rhovz[NO2:-NO2]  - dt/dx[NO2:-NO2]*(BstressZ_x[1:]-BstressZ_x[:-1])
    vx = rhovx/rho
    vy = rhovy/rho
    vz = rhovz/rho   
    
    
    
#      then apply the magnetic flux to evlove the B fields
    by[NO2:-NO2] = by0[NO2:-NO2] - dt/dx[NO2:-NO2]*(by_flux_x[1:]-by_flux_x[:-1])
    bz[NO2:-NO2] = bz0[NO2:-NO2]- dt/dx[NO2:-NO2]*(bz_flux_x[1:]-bz_flux_x[:-1])
 
 #  Update p_per and p_para with evolved B field.
 # The evolved p_para serves as a check parameter 
    bm = np.sqrt(bx**2+by**2+bz**2)
    pper = s1*bm
    ppar = 3*p-2*pper

    

    
    #Outflow boundary
    #:NO2 gives the left boundary cells
    # NO-1:NO2-1:-1 gives the first NO2 active cells in reverse order      
    rho[:NO2] =rho[NO-1:NO2-1:-1]
    vx[:NO2]  = vx[NO-1:NO2-1:-1]
    vy[:NO2]  = vy[NO-1:NO2-1:-1]
    vz[:NO2]  = vz[NO-1:NO2-1:-1]
    p[:NO2]   =  p[NO-1:NO2-1:-1]

    by[:NO2]  = by[NO-1:NO2-1:-1]
    bz[:NO2]  = bz[NO-1:NO2-1:-1]
    pper[:NO2]   =  pper[NO-1:NO2-1:-1]
    ppar[:NO2]   =  ppar[NO-1:NO2-1:-1]


    #-NO2: gives the right boundary cells
    #-NO2-1:-NO-1:-1 gives last NO2 active cells in reverse order
    rho[-NO2:] = rho[-NO2-1:-NO-1:-1]
    p[-NO2:] = p[-NO2-1:-NO-1:-1]
    vx[-NO2:] = vx[-NO2-1:-NO-1:-1]
    vy[-NO2:] = vy[-NO2-1:-NO-1:-1]
    vz[-NO2:] = vz[-NO2-1:-NO-1:-1]

    by[-NO2:] = by[-NO2-1:-NO-1:-1] 
    bz[-NO2:] = bz[-NO2-1:-NO-1:-1]
    pper[-NO2:] = pper[-NO2-1:-NO-1:-1]
    ppar[-NO2:] = ppar[-NO2-1:-NO-1:-1]


    if((step %20)==1):
        plt.clf()
        plt.xlim(-1.5,1.5)
        plt.plot(xc,rho,marker="+",color="slateblue")
        plt.pause(0.01)
        
        
    if Time > 0.2: 
        break

plt.show()